#!/usr/bin/env python3
import click
import os
import platform
import tempfile

from .layout import KeyboardLayout
from .xkb_manager import XKBManager


@click.group()
def cli():
    if platform.system() != 'Linux':
        exit('This command is only compatible with GNU/Linux, sorry.')


@cli.command()
@click.argument('input', nargs=1, type=click.Path(exists=True))
def apply(input):
    """ Apply a Kalamine layout. """

    layout = KeyboardLayout(input)
    f = tempfile.NamedTemporaryFile(mode='w+', suffix='.xkb')
    try:
        f.write(layout.xkb)
        os.system('xkbcomp -w0 %s $DISPLAY' % f.name)
    finally:
        f.close()


@cli.command()
@click.argument('layouts', nargs=-1, type=click.Path(exists=True))
def install(layouts):
    """ Install a list of Kalamine layouts. """
    if len(layouts) == 0:
        return

    xkb = XKBManager()
    for file in layouts:
        xkb.add(KeyboardLayout(file))
    index = xkb.index
    xkb.update()

    print()
    print('Successfully installed. You can try the layout{} with:'.format(
        's' if len(layouts) > 1 else ''
    ))
    for locale, named_layouts in index:
        for name in named_layouts.keys():
            print('    setxkbmap {} -variant {}'.format(locale, name))
    print()


@cli.command(name='list')
@click.argument('mask', default='*')
@click.option('--all', '-a', is_flag=True)
def list_layouts(mask, all):
    """ List all installed Kalamine layouts. """

    xkb = XKBManager()
    list = xkb.list_all(mask) if all else xkb.list(mask)
    for id, desc in sorted(list.items()):
        print('{:<24}   {}'.format(id, desc))


@cli.command()
@click.argument('mask')  # [locale]/[name]
def remove(mask):
    """ Remove an existing Kalamine layout. """

    xkb = XKBManager()
    for layout_id in xkb.list(mask):
        xkb.remove(layout_id)
    xkb.update()



#!/usr/bin/env python3
import os
import unicodedata
import yaml


def _add_spaces_before_combining_chars(text):
    out = ''
    for char in text:
        if unicodedata.combining(char):
            out += ' ' + char
        else:
            out += char
    return out


def _remove_spaces_before_combining_chars(text):
    out = list('')
    for char in text:
        if unicodedata.combining(char):
            out[-1] = char
        else:
            out.append(char)
    return ''.join(out)


def lines_to_text(lines, indent=''):
    out = ''
    for line in lines:
        if len(line):
            out += indent + _add_spaces_before_combining_chars(line)
        out += '\n'
    return out[:-1]


def text_to_lines(text):
    return _remove_spaces_before_combining_chars(text).split('\n')


def open_local_file(file_name):
    return open(os.path.join(os.path.dirname(__file__), file_name))


def load_data(filename):
    return yaml.load(open_local_file(os.path.join('data', filename)))


DEAD_KEYS = load_data('dead_keys.yaml')
LAFAYETTE_KEY = '\u20e1'  # must match the value in dead_keys.yaml
LAYER_KEYS = [
    '- Digits',
    'ae01', 'ae02', 'ae03', 'ae04', 'ae05',
    'ae06', 'ae07', 'ae08', 'ae09', 'ae10',

    '- Letters, first row',
    'ad01', 'ad02', 'ad03', 'ad04', 'ad05',
    'ad06', 'ad07', 'ad08', 'ad09', 'ad10',

    '- Letters, second row',
    'ac01', 'ac02', 'ac03', 'ac04', 'ac05',
    'ac06', 'ac07', 'ac08', 'ac09', 'ac10',

    '- Letters, third row',
    'ab01', 'ab02', 'ab03', 'ab04', 'ab05',
    'ab06', 'ab07', 'ab08', 'ab09', 'ab10',

    '- Pinky keys',
    'ae11', 'ae12', 'ad11', 'ad12', 'ac11',
    'tlde', 'bksl', 'lsgt',

    '- Space bar',
    'spce'
]



#!/usr/bin/env python3
import click
import os

from .layout import KeyboardLayout


def make_all(layout, subdir):
    def out_path(ext=''):
        return os.path.join(subdir, layout.meta['fileName'] + ext)

    if not os.path.exists(subdir):
        os.makedirs(subdir)

    # Windows driver
    klc_path = out_path('.klc')
    open(klc_path, 'w',
         encoding='utf-16le', newline='\r\n').write(layout.klc)
    print('... ' + klc_path)

    # Mac OSX driver
    osx_path = out_path('.keylayout')
    open(osx_path, 'w', newline='\n').write(layout.keylayout)
    print('... ' + osx_path)

    # Linux driver, user-space
    xkb_path = out_path('.xkb')
    open(xkb_path, 'w', newline='\n').write(layout.xkb)
    print('... ' + xkb_path)


@click.command()
@click.argument('input', nargs=-1, type=click.Path(exists=True))
@click.option('--out',
              default='all',
              type=click.Path(),
              help='Keyboard driver(s) to generate.')
def make(input, out):
    """ Convert yaml layout descriptions into OS-specific keyboard layouts. """

    for input_file in input:
        layout = KeyboardLayout(input_file)

        # default: build all in the `dist` subdirectory
        if out == 'all':
            make_all(layout, 'dist')
            continue

        # quick output: reuse the input name and change the file extension
        if out in ['keylayout', 'klc', 'xkb']:
            output_file = os.path.splitext(input_file)[0] + '.' + out
        else:
            output_file = out

        # detailed output
        if output_file.endswith('.klc'):
            open(output_file, 'w',
                 encoding='utf-16le', newline='\r\n').write(layout.klc)
        elif output_file.endswith('.keylayout'):
            open(output_file, 'w', newline='\n').write(layout.keylayout)
        elif output_file.endswith('.xkb'):
            open(output_file, 'w', newline='\n').write(layout.xkb)
        else:
            print('Unsupported output format.')
            return

        # successfully converted, display file name
        print('... ' + output_file)

if __name__ == '__main__':
    make()



#!/usr/bin/env python3
from .utils import load_data, LAYER_KEYS, LAFAYETTE_KEY


##
# Helpers
#

KEY_CODES = load_data('key_codes.yaml')
XKB_KEY_SYM = load_data('key_sym.yaml')


def hex_ord(char):
    return hex(ord(char))[2:].zfill(4)


def xml_proof(char):
    if char not in '<&"\u00a0>':
        return char
    else:
        return '&#x{0};'.format(hex_ord(char))


##
# GNU/Linux: XKB
# - standalone xkb file to be used by `setxkbcomp` (Xorg only)
# - system-wide installer script for Xorg & Wayland
#


def xkb_keymap(layout, eight_levels):
    """ Linux layout. """

    showDescription = True
    maxLength = 16  # `ISO_Level3_Latch` should be the longest symbol name

    output = []
    for keyName in LAYER_KEYS:
        if keyName.startswith('-'):  # separator
            if len(output):
                output.append('')
            output.append('//' + keyName[1:])
            continue

        symbols = []
        description = ' //'
        for layer in layout.layers:
            if keyName in layer:
                symbol = layer[keyName]
                desc = symbol
                if symbol in layout.dead_keys:
                    dk = layout.dead_keys[symbol]
                    desc = dk['alt_self']
                    if dk['char'] == LAFAYETTE_KEY:
                        symbol = 'ISO_Level3_Latch'
                    else:
                        symbol = 'dead_' + dk['name']
                elif symbol in XKB_KEY_SYM \
                        and len(XKB_KEY_SYM[symbol]) <= maxLength:
                    symbol = XKB_KEY_SYM[symbol]
                else:
                    symbol = 'U' + hex_ord(symbol).upper()
            else:
                desc = ' '
                symbol = 'VoidSymbol'

            description += ' ' + desc
            symbols.append(symbol.ljust(maxLength))

        s = 'key <{}> {{[ {}, {}, {}, {}]}};'  # 4-level layout by default
        if layout.has_altgr and layout.has_1dk:
            """ 6 layers are needed: they won't fit on the 4-level format.
            System XKB files require a Neo-like eight-level solution.
            Standalone XKB files work best with a dual-group solution:
            one 4-level group for base+1dk, one two-level group for AltGr.
            """
            if eight_levels:  # system XKB file (patch)
                s = 'key <{}> {{[ {}, {}, {}, {}, {}, {}, {}, {}]}};'
                symbols.append('VoidSymbol'.ljust(maxLength))
                symbols.append('VoidSymbol'.ljust(maxLength))
            else:  # user-space XKB file (standalone)
                s = 'key <{}> {{[ {}, {}, {}, {}],[ {}, {}]}};'
        elif layout.has_altgr:
            del symbols[3]
            del symbols[2]

        line = s.format(* [keyName.upper()] + symbols)
        if showDescription:
            line += description.rstrip()
            if line.endswith('\\'):
                line += ' '  # escape trailing backslash
        output.append(line)

    return output


###
# Windows: KLC
# To be used by the MS Keyboard Layout Creator to generate an installer.
# https://www.microsoft.com/en-us/download/details.aspx?id=22339
# https://levicki.net/articles/2006/09/29/HOWTO_Build_keyboard_layouts_for_Windows_x64.php
# Also supported by KbdEdit: http://www.kbdedit.com/ (non-free).
#


def klc_keymap(layout):
    """ Windows layout, main part. """

    supportedSymbols = \
        '1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

    output = []
    for keyName in LAYER_KEYS:
        if keyName.startswith('-'):
            if len(output):
                output.append('')
            output.append('//' + keyName[1:])
            continue

        symbols = []
        description = '//'
        alpha = False

        for i in [0, 1, 4, 5]:
            layer = layout.layers[i]

            if keyName in layer:
                symbol = layer[keyName]
                desc = symbol
                if symbol in layout.dead_keys:
                    desc = layout.dead_keys[symbol]['alt_space']
                    symbol = hex_ord(desc) + '@'
                else:
                    if i == 0:
                        alpha = symbol.upper() != symbol
                    if symbol not in supportedSymbols:
                        symbol = hex_ord(symbol)
                symbols.append(symbol)
            else:
                desc = ' '
                symbols.append('-1')
            description += ' ' + desc

        if (layout.has_altgr):
            output.append('\t'.join([
                KEY_CODES['klc'][keyName],     # scan code & virtual key
                '1' if alpha else '0',         # affected by CapsLock?
                symbols[0], symbols[1], '-1',  # base layer
                symbols[2], symbols[3],        # altgr layer
                description.strip()
            ]))
        else:
            output.append('\t'.join([
                KEY_CODES['klc'][keyName],     # scan code & virtual key
                '1' if alpha else '0',         # affected by CapsLock?
                symbols[0], symbols[1], '-1',  # base layer
                description.strip()
            ]))

    return output


def klc_deadkeys(layout):
    """ Windows layout, dead keys. """

    output = []

    def appendLine(base, alt):
        s = '{0}\t{1}\t// {2} -> {3}'
        output.append(s.format(hex_ord(base), hex_ord(alt), base, alt))

    for k in layout.dk_index:
        dk = layout.dead_keys[k]

        output.append('// DEADKEY: ' + dk['name'].upper() + ' //{{{')
        output.append('DEADKEY\t' + hex_ord(dk['alt_space']))
        output.append('')

        if k == LAFAYETTE_KEY:
            output.extend(klc_dk_lafayette(layout))
        else:
            for i in range(len(dk['base'])):
                appendLine(dk['base'][i], dk['alt'][i])

        output.append('')
        output.append('// Space bar')
        appendLine('\u00a0', dk['alt_space'])
        appendLine('\u0020', dk['alt_space'])

        output.append('//}}}')
        output.append('')

    return output[:-1]


def klc_dk_index(layout):
    """ Windows layout, dead key index. """

    output = []
    for k in layout.dk_index:
        dk = layout.dead_keys[k]
        output.append('{0}\t"{1}"'.format(hex_ord(dk['alt_space']),
                                          dk['name'].upper()))
    return output


def klc_dk_lafayette(layout):
    """ Windows layout, Lafayette key. """

    output = []
    for i in [0, 1]:
        baseLayer = layout.layers[i]
        extLayer = layout.layers[i + 2]

        for keyName in LAYER_KEYS:
            if keyName.startswith('- Space') or keyName == 'spce':
                continue
            if keyName.startswith('-'):
                if len(output):
                    output.append('')
                output.append('//' + keyName[1:])
                continue
            elif keyName in extLayer:
                base = baseLayer[keyName]
                if base in layout.dead_keys:
                    base = layout.dead_keys[base]['alt_space']
                ext = extLayer[keyName]
                if (ext in layout.dead_keys):
                    ext = layout.dead_keys[ext]['alt_space']
                    lafayette = hex_ord(ext) + '@'
                else:
                    lafayette = hex_ord(ext)

                output.append('\t'.join([
                    hex_ord(base), lafayette, '// ' + base + ' -> ' + ext
                ]))

    return output


###
# MacOS X: keylayout
# https://developer.apple.com/library/content/technotes/tn2056/_index.html
#


def osx_keymap(layout):
    """ Mac OSX layout, main part. """

    str = []
    for index in range(5):
        layer = layout.layers[[0, 1, 0, 4, 5][index]]
        caps = index == 2

        def has_dead_keys(letter):
            for k in layout.dead_keys:
                if letter in layout.dead_keys[k]['base']:
                    return True
            return False

        output = []
        for keyName in LAYER_KEYS:
            if keyName.startswith('-'):
                if len(output):
                    output.append('')
                output.append('<!--' + keyName[1:] + ' -->')
                continue

            symbol = '&#x0010;'
            finalKey = True

            if keyName in layer:
                key = layer[keyName]
                if key in layout.dead_keys:
                    symbol = 'dead_' + layout.dead_keys[key]['name']
                    finalKey = False
                else:
                    symbol = xml_proof(key.upper() if caps else key)
                    finalKey = not has_dead_keys(key)

            c = 'code="{0}"'.format(KEY_CODES['osx'][keyName]).ljust(10)
            a = '{0}="{1}"'.format('output' if finalKey
                                   else 'action', symbol)
            output.append('<key {0} {1} />'.format(c, a))

        str.append(output)
    return str


def osx_actions(layout):
    """ Mac OSX layout, dead key actions. """

    output = []
    deadKeys = []
    dkIndex = []

    def when(state, action):
        s = 'state="{0}"'.format(state).ljust(18)
        if action in layout.dead_keys:
            a = 'next="{0}"'.format(layout.dead_keys[action]['name'])
        elif action.startswith('dead_'):
            a = 'next="{0}"'.format(action[5:])
        else:
            a = 'output="{0}"'.format(xml_proof(action))
        return '  <when {0} {1} />'.format(s, a)

    # spacebar actions
    output.append('<!-- Spacebar -->')
    output.append('<action id="space">')
    output.append(when('none', ' '))
    for k in layout.dk_index:
        dk = layout.dead_keys[k]
        output.append(when(dk['name'], dk['alt_space']))
    output.append('</action>')
    output.append('<action id="nbsp">')
    output.append(when('none', '&#x00a0;'))
    for k in layout.dk_index:
        dk = layout.dead_keys[k]
        output.append(when(dk['name'], dk['alt_space']))
    output.append('</action>')

    # all other actions
    for keyName in LAYER_KEYS:
        if keyName.startswith('-'):
            output.append('')
            output.append('<!--' + keyName[1:] + ' -->')
            continue

        for i in [0, 1]:
            if keyName not in layout.layers[i]:
                continue

            key = layout.layers[i][keyName]
            if i and key == layout.layers[0][keyName]:
                continue
            if key in layout.dead_keys:
                symbol = 'dead_' + layout.dead_keys[key]['name']
            else:
                symbol = xml_proof(key)

            action = []
            for k in layout.dk_index:
                dk = layout.dead_keys[k]
                if key in dk['base']:
                    idx = dk['base'].index(key)
                    action.append(when(dk['name'], dk['alt'][idx]))

            if key in layout.dead_keys:
                deadKeys.append('<action id="{0}">'.format(symbol))
                deadKeys.append(when('none', symbol))
                deadKeys.extend(action)
                deadKeys.append('</action>')
                dkIndex.append(symbol)
            elif len(action):
                output.append('<action id="{0}">'.format(symbol))
                output.append(when('none', symbol))
                output.extend(action)
                output.append('</action>')

        for i in [2, 3, 4, 5]:
            if keyName not in layout.layers[i]:
                continue
            key = layout.layers[i][keyName]
            if key not in layout.dead_keys:
                continue
            symbol = 'dead_' + layout.dead_keys[key]['name']
            if symbol in dkIndex:
                continue
            deadKeys.append('<action id="{0}">'.format(symbol))
            deadKeys.append(when('none', symbol))
            deadKeys.extend(action)
            deadKeys.append('</action>')
            dkIndex.append(symbol)

    return deadKeys + [''] + output


def osx_terminators(layout):
    """ Mac OSX layout, dead key terminators. """

    output = []
    for k in layout.dk_index:
        dk = layout.dead_keys[k]
        s = 'state="{0}"'.format(dk['name']).ljust(18)
        o = 'output="{0}"'.format(xml_proof(dk['alt_self']))
        output.append(' <when {0} {1} />'.format(s, o))
    return output



#!/usr/bin/env python3
import os
import shutil
import sys
import traceback

from lxml import etree
from lxml.builder import E


class XKBManager:
    """ Wrapper to list/add/remove keyboard drivers to XKB. """

    def __init__(self, xkb_root='/usr/share/X11/xkb/'):
        self._rootdir = xkb_root
        self._index = {}

    @property
    def index(self):
        return self._index.items()

    def add(self, layout):
        locale = layout.meta['locale']
        variant = layout.meta['variant']
        if locale not in self._index:
            self._index[locale] = {}
        self._index[locale][variant] = layout

    def remove(self, layout_id):
        locale, variant = layout_id.split('/')
        if locale not in self._index:
            self._index[locale] = {}
        self._index[locale][variant] = None

    def update(self):
        update_symbols(self._rootdir, self._index)  # XKB/symbols/{locales}
        update_rules(self._rootdir, self._index)  # XKB/rules/{base,evdev}.xml
        self._index = {}

    def list(self, mask=''):
        return list_rules(self._rootdir, mask)

    def list_all(self, mask=''):
        return list_rules(self._rootdir, mask, True)


###############################################################################
# Helpers: XKB/symbols
#

""" On GNU/Linux, keyboard layouts must be installed in /usr/share/X11/xkb. To
    be able to revert a layout installation, Kalamine marks layouts like this:

    - XKB/symbols/[locale]: layout definitions
        // KALAMINE::[NAME]::BEGIN
        xkb_symbols "[name]" { ... }
        // KALAMINE::[NAME]::END

    - XKB/rules/{base,evdev}.xml: layout references
        <variant type="kalamine">
            <configItem>
                <name>lafayette42</name>
                <description>French (Lafayette42)</description>
            </configItem>
        </variant>

    Unfortunately, the Lafayette project has released a first installer before
    the XKalamine installer was developed, so we have to handle this situation
    too:

    - XKB/symbols/[locale]: layout definitions
        // LAFAYETTE::BEGIN
        xkb_symbols "lafayette"   { ... }
        xkb_symbols "lafayette42" { ... }
        // LAFAYETTE::END

    - XKB/rules/{base,evdev}.xml: layout references
        <variant type="lafayette">
            <configItem>
                <name>lafayette</name>
                <description>French (Lafayette)</description>
            </configItem>
        </variant>
        <variant type="lafayette">
            <configItem>
                <name>lafayette42</name>
                <description>French (Lafayette42)</description>
            </configItem>
        </variant>

    Consequence: these two Lafayette layouts must be uninstalled together.
    Because of the way they are grouped in symbols/fr, it is impossible to
    remove one without removing the other.
"""

LEGACY_MARK = {
    'begin': '// LAFAYETTE::BEGIN\n',
    'end': '// LAFAYETTE::END\n'
}


def get_symbol_mark(name):
    return {
        'begin': '// KALAMINE::' + name.upper() + '::BEGIN\n',
        'end': '// KALAMINE::' + name.upper() + '::END\n'
    }


def update_symbols_locale(path, named_layouts):
    """ Update Kalamine layouts in an xkb/symbols file. """

    text = ''
    modified_text = False
    NAMES = list(map(lambda n: n.upper(), named_layouts.keys()))

    def is_marked_for_deletion(line):
        if line.startswith('// KALAMINE::'):
            name = line[13:-8]
        elif line.startswith('// LAFAYETTE::'):
            name = 'LAFAYETTE'
        else:
            return False
        return name in NAMES

    with open(path, 'r+') as symbols:

        # look for Kalamine layouts to be updated or removed
        between_marks = False
        closing_mark = ''
        for line in symbols:
            if line.endswith('::BEGIN\n'):
                if is_marked_for_deletion(line):
                    closing_mark = line[:-6] + 'END\n'
                    modified_text = True
                    between_marks = True
                    text = text.rstrip()
                else:
                    text += line
            elif line.endswith('::END\n'):
                if between_marks and line.startswith(closing_mark):
                    between_marks = False
                    closing_mark = ''
                else:
                    text += line
            elif not between_marks:
                text += line

        # clear previous Kalamine layouts if needed
        if modified_text:
            symbols.seek(0)
            symbols.write(text.rstrip() + '\n')
            symbols.truncate()

        # add new Kalamine layouts
        for name, layout in named_layouts.items():
            if layout is None:
                print('      - ' + name)
            else:
                print('      + ' + name)
                MARK = get_symbol_mark(name)
                symbols.write('\n')
                symbols.write(MARK['begin'])
                symbols.write(layout.xkb_patch.rstrip() + '\n')
                symbols.write(MARK['end'])

        symbols.close()


def update_symbols(xkb_root, kbindex):
    """ Update Kalamine layouts in all xkb/symbols files. """

    for locale, named_layouts in kbindex.items():
        path = os.path.join(xkb_root, 'symbols', locale)
        if not os.path.exists(path):
            exit_LocaleNotSupported(locale)

        try:
            if not os.path.isfile(path + '.orig'):
                # backup, just in case :-)
                shutil.copy(path, path + '.orig')
                print('... ' + path + '.orig (backup)')

            print('... ' + path)
            update_symbols_locale(path, named_layouts)

        except Exception as e:
            exit_FileNotWritable(e, path)


###############################################################################
# Helpers: XKB/rules
#

def get_rules_locale(tree, locale):
    query = '//layout/configItem/name[text()="%s"]/../..' % locale
    result = tree.xpath(query)
    if len(result) != 1:
        exit_LocaleNotSupported(locale)
    return tree.xpath(query)[0]


def remove_rules_variant(variant_list, name):
    signatures = ['kalamine']
    if name.lower().startswith('lafayette'):
        signatures.append('lafayette')

    for signature in signatures:
        query = 'variant[@type="{}"]/configItem/name[text()="{}"]/../..'.\
                format(signature, name)
        for variant in variant_list.xpath(query):
            variant.getparent().remove(variant)


def add_rules_variant(variant_list, name, description):
    variant_list.append(
        E.variant(
            E.configItem(
                E.name(name),
                E.description(description)
            ), type='kalamine'))


def update_rules(xkb_root, kbindex):
    """ Update references in XKB/rules/{base,evdev}.xml. """

    for filename in ['base.xml', 'evdev.xml']:
        try:
            path = os.path.join(xkb_root, 'rules', filename)
            tree = etree.parse(path, etree.XMLParser(remove_blank_text=True))

            for locale, named_layouts in kbindex.items():
                vlist = get_rules_locale(tree, locale).xpath('variantList')
                if len(vlist) != 1:
                    exit('Error: unexpected xml format in %s.' % path)
                for name, layout in named_layouts.items():
                    remove_rules_variant(vlist[0], name)
                    if layout is not None:
                        description = layout.meta['description']
                        add_rules_variant(vlist[0], name, description)

            tree.write(path, pretty_print=True, xml_declaration=True,
                       encoding='utf-8')
            print('... ' + path)

        except Exception as e:
            exit_FileNotWritable(e, path)


def list_rules(xkb_root, mask='', include_non_kalamine_variants=False):
    """ List all installed Kalamine layouts. """

    def matches(string, mask):
        return mask == '*' or mask == string

    if mask == '' or mask == '*':
        locale_mask = '*'
        variant_mask = '*'
    else:
        m = mask.split('/')
        if len(m) != 2:
            exit('Error: expecting a [locale]/[variant] mask.')
        locale_mask, variant_mask = m

    query = '//variant'
    if not include_non_kalamine_variants:
        query += '[@type]'

    layouts = {}
    for filename in ['base.xml', 'evdev.xml']:
        tree = etree.parse(os.path.join(xkb_root, 'rules', filename))
        for variant in tree.xpath(query):
            locale = variant.xpath('../../configItem/name')[0].text
            name = variant.xpath('configItem/name')[0].text
            desc = variant.xpath('configItem/description')[0].text
            id = locale + '/' + name
            if id not in layouts \
               and matches(locale, locale_mask) \
               and matches(name, variant_mask):
                layouts[id] = desc

    return layouts


###############################################################################
# Exception Handling (there must be a better way...)
#

def exit(message):
    print('')
    print(message)
    sys.exit(1)


def exit_LocaleNotSupported(locale):
    exit('Error: the `%s` locale is not supported.' % locale)


def exit_FileNotWritable(exception, path):
    if isinstance(exception, PermissionError):  # noqa: F821
        exit('Permission denied. Are you root?')
    elif isinstance(exception, IOError):
        exit('Error: could not write to file %s.' % path)
    else:  # exit('Unexpected error: ' + sys.exc_info()[0])
        exit('Error: {}.\n{}'.format(exception, traceback.format_exc()))
